@@ -1,6 +1,6 @@
 use v6;
 
-unit module TAP:ver<0.3.1>;
+unit module TAP:ver<0.3.3>;
 
 role Entry {
     has Str:D $.raw is required handles <Str>;
@@ -389,6 +389,8 @@ role Formatter {
     has Bool:D $.ignore-exit = False;
 }
 role Reporter {
+    has Output:D $.output is required;
+    has Formatter:D $.formatter handles<volume> is required;
     method summarize(TAP::Aggregator, Bool $interrupted, Duration $duration) { ... }
     method open-test(Str $) { ... }
 }
@@ -529,13 +531,6 @@ class Formatter::Text does Formatter {
     }
 }
 class Reporter::Text does Reporter {
-    has Output $!output;
-    has Formatter::Text $!formatter handles <volume>;
-
-    submethod BUILD(:@names, :$!output = Output::Handle.new, :$volume = Normal, :$timer = False, Bool :$ignore-exit) {
-        $!formatter = Formatter::Text.new(:@names, :$volume, :$timer, :$ignore-exit);
-    }
-
     method open-test(Str $name) {
         my $header = $!formatter.format-name($name);
         Reporter::Text::Session.new(:$name, :$header, :reporter(self));
@@ -551,9 +546,8 @@ class Reporter::Text does Reporter {
     }
 }
 
-class Formatter::Console is Formatter::Text {
-    has $.color;
-    has &colored = $!color && (try require Terminal::ANSIColor) !=== Nil
+class Formatter::Color is Formatter::Text {
+    has &colored = (try require Terminal::ANSIColor) !=== Nil
         ?? ::('Terminal::ANSIColor::EXPORT::DEFAULT::&colored')
         !! sub ($text, $) { $text };
     method format-success(Str $output) {
@@ -596,20 +590,13 @@ class Reporter::Console::Session does Session {
     }
 }
 class Reporter::Console does Reporter {
-    has Formatter::Console $!formatter handles <volume>;
-    has Int $!lastlength;
-    has Supplier $events;
+    has Int $!lastlength = 0;
+    has Supplier $events = Supplier.new;
     has Reporter::Console::Session @!active;
-    has Int $!tests;
-    has Int $!fails;
-
-    submethod BUILD(:@names, Output :$output = Output::Handle.new, :$volume = Normal, :$timer = False, Bool :$ignore-exit = False, Bool :$color) {
-        $!formatter = Formatter::Console.new(:@names, :$volume, :$timer, :$ignore-exit, :$color);
-        $!lastlength = 0;
-        $!events = Supplier.new;
-        @!active .= new;
-        $!tests = 0;
+    has Int $!tests = 0;
+    has Int $!fails = 0;
 
+    submethod TWEAK(:$!output, :$!formatter) {
         my $now = 0;
         my $start = now;
 
@@ -621,13 +608,13 @@ class Reporter::Console does Reporter {
             my @items = @!active.map(*.summary);
             my $ruler = ($header, |@items).join('  ') ~ ')===';
             $ruler = $ruler.substr(0,70) if $ruler.chars > 70;
-            $output.print($!formatter.format-return($ruler));
+            $!output.print($!formatter.format-return($ruler));
         }
         multi receive('update', Str $name, Str $header, Int $number, Int $plan) {
             return if self.volume <= Quiet;
             if @!active.elems == 1 {
                 my $status = ($header, $number, '/', $plan // '?').join('');
-                $output.print($!formatter.format-return($status));
+                $!output.print($!formatter.format-return($status));
                 $!lastlength = $status.chars + 1;
             }
             else {
@@ -636,16 +623,16 @@ class Reporter::Console does Reporter {
         }
         multi receive('bailout', Str $explanation) {
             return if self.volume <= Quiet;
-            $output.print($!formatter.format-failure("Bailout called.  Further testing stopped: $explanation\n"));
+            $!output.print($!formatter.format-failure("Bailout called.  Further testing stopped: $explanation\n"));
         }
         multi receive('result', Reporter::Console::Session $session, TAP::Result $result) {
             return if self.volume <= Quiet;
-            $output.print($!formatter.format-return(' ' x $!lastlength) ~ $!formatter.format-result($session, $result));
+            $!output.print($!formatter.format-return(' ' x $!lastlength) ~ $!formatter.format-result($session, $result));
             @!active = @!active.grep(* !=== $session);
             output-ruler(True) if @!active.elems > 1;
         }
         multi receive('summary', TAP::Aggregator $aggregator, Bool $interrupted, Duration $duration) {
-            $output.print($!formatter.format-summary($aggregator, $interrupted, $duration)) unless self.volume == Silent;
+            $!output.print($!formatter.format-summary($aggregator, $interrupted, $duration)) unless self.volume == Silent;
         }
 
         $!events.Supply.act(-> @args { receive(|@args) });
@@ -808,9 +795,10 @@ class Source::Proc does Source {
     has Str $.path is required;
     has @.args;
     has $.err is required;
+    has $.cwd is required;
 }
 class Source::File does Source {
-    has Str $.filename;
+    has IO(Str) $.filename handles<slurp>;
 }
 class Source::String does Source {
     has Str $.content;
@@ -857,7 +845,7 @@ class Async {
                 die "Unknown error handler";
             }
         }
-        my $process = $async.start;
+        my $process = $async.start(:cwd($proc.cwd));
         my $start-time = now;
         my $timer = $process.then({ now - $start-time });
         Run.new(:$process, :killer($async), :$timer, :$events);
@@ -868,7 +856,7 @@ class Async {
         Run.new(:$events);
     }
     multi get_runner(Source::File $file) {
-        my $events = parser(supply { emit $file.filename.IO.slurp(:close) });
+        my $events = parser(supply { emit $file.slurp(:close) });
         Run.new(:$events);
     }
     multi get_runner(Source::String $string) {
@@ -896,15 +884,17 @@ class Harness {
         method make-source {...};
     }
     role SourceHandler::Proc does SourceHandler {
-        has Str:D $.path is required;
-        has @.args;
-        method make-source(Str:D $name, Any:D :$err) {
-            TAP::Source::Proc.new(:$name, :$!path, :args[ |@!args, $name ], :$err);
-        }
+    }
+    my sub normalize-path($path, IO::Path $cwd) {
+        $path ~~ IO ?? $path.IO.relative($cwd) !! ~$path
     }
     class SourceHandler::Raku does SourceHandler::Proc {
-        submethod BUILD(:@incdirs, Str:D :$!path = $*EXECUTABLE.absolute) {
-            @!args = @incdirs.map("-I" ~ *);
+        has Str:D $.path = $*EXECUTABLE.absolute;
+        has Str @.incdirs;
+        method make-source(Str:D $name, Any:D :$err, IO::Path:D :$cwd, :@include-dirs = (), *%) {
+            my @dirs = map { normalize-path($_, $cwd) }, |@include-dirs, |@!incdirs;
+            my @args = |@dirs.map("-I" ~ *), $name;
+            TAP::Source::Proc.new(:$name, :$!path, :@args, :$err, :$cwd);
         }
         method can-handle(Str $name) {
             0.5;
@@ -914,12 +904,15 @@ class Harness {
         # This may later give deprecation warnings
     }
     class SourceHandler::Exec does SourceHandler::Proc {
-        method new (*@ ($path, *@args)) {
-            self.bless(:$path, :@args);
-        }
+        has Str:D $.path is required;
+        has @.args;
         method can-handle(Str $name) {
             1;
         }
+        method make-source(Str:D $name, Any:D :$err, IO::Path:D :$cwd, *%) {
+            my @args = |@!args, $name;
+            TAP::Source::Proc.new(:$name, :$!path, :@args, :$err, :$cwd);
+        }
     }
 
     my sub load-reporter(Str $name --> Reporter) {
@@ -931,19 +924,13 @@ class Harness {
     }
 
     subset OutVal where any(IO::Handle:D, Supplier:D);
-    my multi is-terminal(IO::Handle:D $out) {
-        return $out.t;
-    }
-    my multi is-terminal(Supplier:D $out) {
-        return False
-    }
 
     has SourceHandler @.handlers = SourceHandler::Raku.new();
     has IO::Handle $.handle = $*OUT;
     has OutVal $.output = $!handle;
     has Formatter::Volume $.volume = ?%*ENV<HARNESS_VERBOSE> ?? Verbose !! Normal;
     has Str %!env-options = (%*ENV<HARNESS_OPTIONS> // '').split(':').grep(*.chars).map: { / ^ (.) (.*) $ /; ~$0 => val(~$1) };
-    has TAP::Reporter:U $.reporter-class = %!env-options<r> ?? load-reporter(%!env-options<r>) !! is-terminal($!output) && $!volume < Verbose ?? TAP::Reporter::Console !! TAP::Reporter::Text;
+    has TAP::Reporter:U $.reporter-class;
     has Int:D $.jobs = %!env-options<j> // 1;
     has Bool:D $.timer = ?%*ENV<HARNESS_TIMER>;
     subset ErrValue where any(IO::Handle:D, Supplier, 'stderr', 'ignore', 'merge');
@@ -951,8 +938,7 @@ class Harness {
     has Bool:D $.ignore-exit = ?%*ENV<HARNESS_INGORE_EXIT>;
     has Bool:D $.trap = False;
     has Bool:D $.loose = $*PERL.compiler.version before 2017.09;
-    my @safe-terminals = <xterm eterm vte konsole color>;
-    has Bool:D $.color= so %!env-options<c> // %*ENV<HARNESS_COLOR> // !%*ENV<NO_COLOR>.defined && is-terminal($!output) && (%*ENV<TERM> // '') ~~ / @safe-terminals /;
+    has Bool $.color;
 
     class Run {
         has Promise $.waiter handles <result>;
@@ -972,9 +958,6 @@ class Harness {
             $events.act({ $output.say(~$^entry) }, :done({ $output.flush }), :quit({ $output.flush }));
         }
     }
-    method make-source(Str $name) {
-        @!handlers.max(*.can-handle($name)).make-source($name, :$!err);
-    }
     my &sigint = sub { signal(SIGINT) }
 
     my multi make-output(IO::Handle:D $handle) {
@@ -983,11 +966,46 @@ class Harness {
     my multi make-output(Supplier:D $supplier) {
         return Output::Supplier.new(:$supplier);
     }
-    method run(*@names) {
+
+    my sub get-reporter(Reporter $reporter-class, %env-options, $output, $volume) {
+        if $reporter-class !=== Reporter {
+            return $reporter-class;
+        } elsif %env-options<r> {
+            return load-reporter(%env-options<r>);
+        } elsif $output.terminal && $volume < Verbose {
+            return TAP::Reporter::Console;
+        } else {
+            return TAP::Reporter::Text;
+        }
+    }
+
+    my sub get-color(Bool $color, %env-options, Output $output) {
+        with $color {
+            return $color;
+        }
+        orwith %env-options<c> {
+            return True;
+        }
+        orwith %*ENV<HARNESS_COLOR> {
+            return ?%*ENV<HARNESS_COLOR>;
+        }
+        orwith %*ENV<NO_COLOR> {
+            return False;
+        }
+        else {
+            state @safe-terminals = <xterm eterm vte konsole color>;
+            return $output.terminal && (%*ENV<TERM> // '') ~~ / @safe-terminals /;
+        }
+    }
+
+    method run(*@names, IO(Str) :$cwd = $*CWD, OutVal :$out = $!output, ErrValue :$err = $!err, *%handler-args) {
         my $killed = Promise.new;
         my $aggregator = self.make-aggregator;
-        my $output = make-output($!output);
-        my $reporter = $!reporter-class.new(:@names, :$!timer, :$!ignore-exit, :$!volume, :$output, :$!color);
+        my $output = make-output($out);
+        my $formatter-class = get-color($!color, %!env-options, $output) ?? Formatter::Color !! Formatter::Text;
+        my $formatter = $formatter-class.new(:@names, :$!volume, :$!timer, :$!ignore-exit);
+        my $reporter-class = get-reporter($!reporter-class, %!env-options, $output, $!volume);
+        my $reporter = $reporter-class.new(:$output, :$formatter);
 
         my @working;
         my $waiter = start {
@@ -995,8 +1013,9 @@ class Harness {
             my $begin = now;
             try {
                 for @names -> $name {
-                    my $session = $reporter.open-test($name);
-                    my $source = self.make-source($name);
+                    my $path = normalize-path($name, $cwd);
+                    my $session = $reporter.open-test($path);
+                    my $source = @!handlers.max(*.can-handle($name)).make-source($path, :$err, :$cwd, |%handler-args);
                     my $parser = TAP::Async.new(:$source, :$killed, :$!loose);
                     $session.listen($parser.events);
                     self.add-handlers($parser.events, $output);
